import os
import ctypes
from ctypes import wintypes
import time
import msal
from ._cache_lock import CrossPlatLock

_local_free = ctypes.windll.kernel32.LocalFree
_memcpy = ctypes.cdll.msvcrt.memcpy
_crypt_protect_data = ctypes.windll.crypt32.CryptProtectData
_crypt_unprotect_data = ctypes.windll.crypt32.CryptUnprotectData
_CRYPTPROTECT_UI_FORBIDDEN = 0x01


class DATA_BLOB(ctypes.Structure):
    """
    A wrapper for interacting with the _CRYPTOAPI_BLOB type and its many aliases. This type is exposed from Wincrypt.h
    in XP and above.

    See documentation for this type at:
    https://msdn.microsoft.com/en-us/7a06eae5-96d8-4ece-98cb-cf0710d2ddbd
    """
    _fields_ = [("cbData", wintypes.DWORD), ("pbData", ctypes.POINTER(ctypes.c_char))]

    def raw(self):
        # type: () -> bytes
        cb_data = int(self.cbData)
        pb_data = self.pbData
        buffer = ctypes.c_buffer(cb_data)
        _memcpy(buffer, pb_data, cb_data)
        _local_free(pb_data)
        return buffer.raw


# This code is modeled from a StackOverflow question, which can be found here:
# https://stackoverflow.com/questions/463832/using-dpapi-with-python
class WindowsDataProtectionAgent(object):
    def __init__(self, **kwargs):
        # type: (str) -> None
        if 'entropy' in kwargs:
            entropy = kwargs['entropy'].encode('utf-8')
            self._entropy_buffer = ctypes.c_buffer(entropy, len(entropy))
            self._entropy_blob = DATA_BLOB(len(entropy), self._entropy_buffer)
        else:
            self._entropy_buffer = None
            self._entropy_blob = None

    def protect(self, message):
        # type: (str) -> bytes

        message = message.encode('utf-8')
        message_buffer = ctypes.c_buffer(message, len(message))
        message_blob = DATA_BLOB(len(message), message_buffer)
        result = DATA_BLOB()

        if self._entropy_blob:
            entropy = ctypes.byref(self._entropy_blob)
        else:
            entropy = None

        if _crypt_protect_data(
                ctypes.byref(message_blob),
                u"python_data",
                entropy,
                None,
                None,
                _CRYPTPROTECT_UI_FORBIDDEN,
                ctypes.byref(result)):
            return result.raw()
        return b''

    def unprotect(self, cipher_text):
        # type: (bytes) -> str
        ct_buffer = ctypes.c_buffer(cipher_text, len(cipher_text))
        ct_blob = DATA_BLOB(len(cipher_text), ct_buffer)
        result = DATA_BLOB()

        if self._entropy_blob:
            entropy = ctypes.byref(self._entropy_blob)
        else:
            entropy = None

        if _crypt_unprotect_data(
            ctypes.byref(ct_blob),
            None,
            entropy,
            None,
            None,
            _CRYPTPROTECT_UI_FORBIDDEN,
            ctypes.byref(result)
        ):
            return result.raw().decode('utf-8')
        return u''


class _WindowsTokenCache(msal.SerializableTokenCache):
    DEFAULT_CACHE_LOCATION = os.path.join(os.getenv('LOCALAPPDATA'), '.IdentityService', 'msal.cache')
    DEFAULT_ENTROPY = ''

    def __init__(self, **kwargs):
        super(_WindowsTokenCache, self).__init__()

        self._cache_location = _WindowsTokenCache.DEFAULT_CACHE_LOCATION  # type: str
        if 'cache_location' in kwargs:
            self._cache_location = kwargs['cache_location'] or _WindowsTokenCache.DEFAULT_CACHE_LOCATION
        self._lock_location = self._cache_location + '.lockfile'

        entropy = _WindowsTokenCache.DEFAULT_ENTROPY
        if 'entropy' in kwargs:
            entropy = kwargs['entropy']
        self._dp_agent = WindowsDataProtectionAgent(entropy=entropy)
        self._last_sync = 0  # _last_sync is a Unixtime

    def _has_state_changed(self):
        # type: () -> Bool
        try:
            return self.has_state_changed or self._last_sync < os.path.getmtime(self._cache_location)
        except FileNotFoundError:
            return False

    def add(self, event, **kwargs):
        if self._has_state_changed:
            try:
                self._read()
            except FileNotFoundError:
                pass
        super(_WindowsTokenCache, self).add(event, **kwargs)
        self._write()

    def update_rt(self, rt_item, new_rt):
        if self.has_state_changed:
            try:
                self._read()
            except FileNotFoundError:
                pass
        super(_WindowsTokenCache, self).update_rt(rt_item, new_rt)
        self._write()

    def remove_rt(self, rt_item):
        if self._has_state_changed:
            try:
                self._read()
            except FileNotFoundError:
                pass
        super(_WindowsTokenCache, self).remove_rt(rt_item)
        self._write()

    def find(self, credential_type, target=None, query=None):
        if self._has_state_changed:
            try:
                self._read()
            except FileNotFoundError:
                pass
        return super(_WindowsTokenCache, self).find(credential_type, target=target, query=query)

    def _write(self):
        with CrossPlatLock(self._lock_location), open(self._cache_location, 'wb') as fh:
            fh.write(self._dp_agent.protect(self.serialize()))
        self._last_sync = int(time.time())

    def _read(self):
        with CrossPlatLock(self._lock_location), open(self._cache_location, 'rb') as fh:
            cipher_text = fh.read()
            contents = self._dp_agent.unprotect(cipher_text)
            self.deserialize(contents)
        self._last_sync = int(time.time())



import os
import datetime
import time
import psutil
import portalocker


class CrossPlatLock(object):
    TIMEOUT = datetime.timedelta(minutes=1)
    RETRY_WAIT = datetime.timedelta(milliseconds=100)
    RETRY_COUNT = int(TIMEOUT / RETRY_WAIT)

    def __init__(self, lockfile_path):
        self._lockpath = lockfile_path

    def __enter__(self):
        pid = os.getpid()
        proc = psutil.Process(pid)
        lock_dir = os.path.dirname(self._lockpath)
        os.makedirs(lock_dir, exist_ok=True)

        self._fh = open(self._lockpath, 'wb+', buffering=0)
        portalocker.lock(self._fh, portalocker.LOCK_EX)
        self._fh.write('{} {}'.format(pid, proc.name()).encode('utf-8'))

    def __exit__(self, *args):
        self._fh.close()
        try:
            os.remove(self._lockpath)
        except PermissionError:
            pass



import os
import ctypes as _ctypes
import time
import msal

OS_result = _ctypes.c_int32


class KeychainError(OSError):
    def __init__(self, exit_status):
        self.exit_status = exit_status
        # TODO: use SecCopyErrorMessageString to fetch the appropriate message here.
        self.message = '{} ' \
                       'check https://opensource.apple.com/source/CarbonHeaders/CarbonHeaders-18.1/MacErrors.h'.format(
                            self.exit_status)


class KeychainAccessDeniedError(KeychainError):
    EXIT_STATUS = -128

    def __init__(self):
        super().__init__(KeychainAccessDeniedError.EXIT_STATUS)


class NoSuchKeychainError(KeychainError):
    EXIT_STATUS = -25294

    def __init__(self, name):
        super().__init__(NoSuchKeychainError.EXIT_STATUS)
        self.name = name


class NoDefaultKeychainError(KeychainError):
    EXIT_STATUS = -25307

    def __init__(self):
        super().__init__(NoDefaultKeychainError.EXIT_STATUS)


class KeychainItemNotFoundError(KeychainError):
    EXIT_STATUS = -25300

    def __init__(self, service_name, account_name):
        super().__init__(KeychainItemNotFoundError.EXIT_STATUS)
        self.service_name = service_name
        self.account_name = account_name


def _construct_error(exit_status, **kwargs):
    if exit_status == KeychainAccessDeniedError.EXIT_STATUS:
        return KeychainAccessDeniedError()
    if exit_status == NoSuchKeychainError.EXIT_STATUS:
        return NoSuchKeychainError(**kwargs)
    if exit_status == NoDefaultKeychainError.EXIT_STATUS:
        return NoDefaultKeychainError()
    if exit_status == KeychainItemNotFoundError.EXIT_STATUS:
        return KeychainItemNotFoundError(**kwargs)


def _get_native_location(name):
    # type: (str) -> str
    """
    Fetches the location of a native MacOS library.
    :param name: The name of the library to be loaded.
    :return: The location of the library on a MacOS filesystem.
    """
    return '/System/Library/Frameworks/{0}.framework/{0}'.format(name)


# Load native MacOS libraries
try:
    _security = _ctypes.CDLL(_get_native_location('Security'))
    _core = _ctypes.CDLL(_get_native_location('CoreFoundation'))
except OSError:
    _security = None
    _core = None

# Bind CFRelease from native MacOS libraries.
_coreRelease = _core.CFRelease
_coreRelease.argtypes = (
    _ctypes.c_void_p,
)

# Bind SecCopyErrorMessageString from native MacOS libraries.
# https://developer.apple.com/documentation/security/1394686-seccopyerrormessagestring?language=objc
_securityCopyErrorMessageString = _security.SecCopyErrorMessageString
_securityCopyErrorMessageString.argtypes = (
    OS_result,
    _ctypes.c_void_p
)
_securityCopyErrorMessageString.restype = _ctypes.c_char_p

# Bind SecKeychainOpen from native MacOS libraries.
# https://developer.apple.com/documentation/security/1396431-seckeychainopen
_securityKeychainOpen = _security.SecKeychainOpen
_securityKeychainOpen.argtypes = (
    _ctypes.c_char_p,
    _ctypes.POINTER(_ctypes.c_void_p)
)
_securityKeychainOpen.restype = OS_result

# Bind SecKeychainCopyDefault from native MacOS libraries.
# https://developer.apple.com/documentation/security/1400743-seckeychaincopydefault?language=objc
_securityKeychainCopyDefault = _security.SecKeychainCopyDefault
_securityKeychainCopyDefault.argtypes = (
    _ctypes.POINTER(_ctypes.c_void_p),
)
_securityKeychainCopyDefault.restype = OS_result


# Bind SecKeychainItemFreeContent from native MacOS libraries.
_securityKeychainItemFreeContent = _security.SecKeychainItemFreeContent
_securityKeychainItemFreeContent.argtypes = (
    _ctypes.c_void_p,
    _ctypes.c_void_p,
)
_securityKeychainItemFreeContent.restype = OS_result

# Bind SecKeychainItemModifyAttributesAndData from native MacOS libraries.
_securityKeychainItemModifyAttributesAndData = _security.SecKeychainItemModifyAttributesAndData
_securityKeychainItemModifyAttributesAndData.argtypes = (
    _ctypes.c_void_p,
    _ctypes.c_void_p,
    _ctypes.c_uint32,
    _ctypes.c_void_p,
)
_securityKeychainItemModifyAttributesAndData.restype = OS_result

# Bind SecKeychainFindGenericPassword from native MacOS libraries.
# https://developer.apple.com/documentation/security/1397301-seckeychainfindgenericpassword?language=objc
_securityKeychainFindGenericPassword = _security.SecKeychainFindGenericPassword
_securityKeychainFindGenericPassword.argtypes = (
    _ctypes.c_void_p,
    _ctypes.c_uint32,
    _ctypes.c_char_p,
    _ctypes.c_uint32,
    _ctypes.c_char_p,
    _ctypes.POINTER(_ctypes.c_uint32),
    _ctypes.POINTER(_ctypes.c_void_p),
    _ctypes.POINTER(_ctypes.c_void_p),
)
_securityKeychainFindGenericPassword.restype = OS_result
# Bind SecKeychainAddGenericPassword from native MacOS
# https://developer.apple.com/documentation/security/1398366-seckeychainaddgenericpassword?language=objc
_securityKeychainAddGenericPassword = _security.SecKeychainAddGenericPassword
_securityKeychainAddGenericPassword.argtypes = (
    _ctypes.c_void_p,
    _ctypes.c_uint32,
    _ctypes.c_char_p,
    _ctypes.c_uint32,
    _ctypes.c_char_p,
    _ctypes.c_uint32,
    _ctypes.c_char_p,
    _ctypes.POINTER(_ctypes.c_void_p),
)
_securityKeychainAddGenericPassword.restype = OS_result


class Keychain(object):
    """Encapsulates the interactions with a particular MacOS Keychain."""
    def __init__(self, filename=None):
        # type: (str) -> None
        self._ref = _ctypes.c_void_p()

        if filename:
            filename = os.path.expanduser(filename)
            self._filename = filename.encode('utf-8')
        else:
            self._filename = None

    def __enter__(self):
        if self._filename:
            status = _securityKeychainOpen(self._filename, self._ref)
        else:
            status = _securityKeychainCopyDefault(self._ref)

        if status:
            raise OSError(status)
        return self

    def __exit__(self, *args):
        if self._ref:
            _coreRelease(self._ref)

    def get_generic_password(self, service, account_name):
        # type: (str, str) -> str
        service = service.encode('utf-8')
        account_name = account_name.encode('utf-8')

        length = _ctypes.c_uint32()
        contents = _ctypes.c_void_p()
        exit_status = _securityKeychainFindGenericPassword(
            self._ref,
            len(service),
            service,
            len(account_name),
            account_name,
            length,
            contents,
            None,
        )

        if exit_status:
            raise _construct_error(exit_status=exit_status, service_name=service, account_name=account_name)

        value = _ctypes.create_string_buffer(length.value)
        _ctypes.memmove(value, contents.value, length.value)
        _securityKeychainItemFreeContent(None, contents)
        return value.raw.decode('utf-8')

    def set_generic_password(self, service, account_name, value):
        # type: (str, str, str) -> None
        service = service.encode('utf-8')
        account_name = account_name.encode('utf-8')
        value = value.encode('utf-8')

        entry = _ctypes.c_void_p()
        find_exit_status = _securityKeychainFindGenericPassword(
            self._ref,
            len(service),
            service,
            len(account_name),
            account_name,
            None,
            None,
            entry,
        )

        if not find_exit_status:
            modify_exit_status = _securityKeychainItemModifyAttributesAndData(
                entry,
                None,
                len(value),
                value,
            )
            if modify_exit_status:
                raise _construct_error(modify_exit_status, service_name=service, account_name=account_name)

        elif find_exit_status == KeychainItemNotFoundError.EXIT_STATUS:
            add_exit_status = _securityKeychainAddGenericPassword(
                self._ref,
                len(service),
                service,
                len(account_name),
                account_name,
                len(value),
                value,
                None
            )

            if add_exit_status:
                raise _construct_error(add_exit_status, service_name=service, account_name=account_name)
        else:
            raise _construct_error(find_exit_status, service_name=service, account_name=account_name)

    def get_internet_password(self, service, username):
        # type: (str, str) -> str
        raise NotImplementedError()

    def set_internet_password(self, service, username, value):
        # type: (str, str, str) -> str
        raise NotImplementedError()


class _OSXTokenCache(msal.SerializableTokenCache):
    DEFAULT_SERVICE_NAME = 'Microsoft.Developer.IdentityService'
    DEFAULT_ACCOUNT_NAME = 'MSALCache'

    def __init__(self, **kwargs):
        super(_OSXTokenCache, self).__init__()
        self._service_name = kwargs.get('service_name', _OSXTokenCache.DEFAULT_SERVICE_NAME)
        self._account_name = kwargs.get('account_name', _OSXTokenCache.DEFAULT_ACCOUNT_NAME)

    def add(self, event, **kwargs):
        super(_OSXTokenCache, self).add(event, **kwargs)
        self._write()

    def update_rt(self, rt_item, new_rt):
        super(_OSXTokenCache, self).update_rt(rt_item, new_rt)
        self._write()

    def remove_rt(self, rt_item):
        super(_OSXTokenCache, self).remove_rt(rt_item)
        self._write()

    def find(self, credential_type, target=None, query=None):
        self._read()
        return super(_OSXTokenCache, self).find(credential_type, target=target, query=query)

    def _read(self):
        with Keychain() as locker:
            contents = locker.get_generic_password(self._service_name, self._account_name)
        self.deserialize(contents)
        self._last_sync = int(time.time())

    def _write(self):
        with Keychain() as locker:
            locker.set_generic_password(self._service_name, self._account_name, self.serialize())
        self._last_sync = int(time.time())



__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from .token_provider import TokenProvider, TokenProviderChain, SharedTokenCacheProvider, DEFAULT_TOKEN_CHAIN
from .token_cache import FileTokenCache, get_protected_token_cache



import sys
import os
import tempfile
import time
import msal

if sys.platform.startswith('win'):
    from ._windows import _WindowsTokenCache
elif sys.platform.startswith('darwin'):
    from ._osx import _OSXTokenCache


def get_protected_token_cache(enforce_encryption=False, **kwargs):
    if sys.platform.startswith('win'):
        return _WindowsTokenCache(**kwargs)
    elif sys.platform.startswith('darwin'):
        return _OSXTokenCache(**kwargs)
    elif enforce_encryption:
        raise RuntimeError('no protected token cache for platform {}'.format(sys.platform))
    else:
        return FileTokenCache(**kwargs)


class FileTokenCache(msal.SerializableTokenCache):
    # TODO: Find correct location for this file
    DEFAULT_FILE_LOCATION = os.path.join(tempfile.gettempdir(), "msal.cache.txt")

    def __init__(self, **kwargs):
        self._file_location = kwargs.get('cache_location', FileTokenCache.DEFAULT_FILE_LOCATION)
        self._last_sync = 0

    def add(self, event, **kwargs):
        super(FileTokenCache, self).add(event, **kwargs)
        self._write()

    def update_rt(self, rt_item, new_rt):
        super(FileTokenCache, self).update_rt(rt_item, new_rt)
        self._write()

    def remove_rt(self, rt_item):
        super(FileTokenCache, self).remove_rt(rt_item)
        self._write()

    def find(self, credential_type, target=None, query=None):
        if self._has_state_changed():
            self._read()
        return super(_WindowsTokenCache, self).find(credential_type, target=target, query=query)

    def _write(self):
        with open(self._file_location, "w") as fh:
            fh.write(self.serialize())
        self._last_sync = int(time.time())

    def _read(self):
        with open(self._file_location, "r") as fh:
            self.deserialize(fh.read())
        self._last_sync = int(time.time())



import os
import abc
from msal.application import PublicClientApplication, ConfidentialClientApplication
from .token_cache import get_protected_token_cache

_DEFAULT_CLIENT_ID = '04b07795-8ddb-461a-bbee-02f9e1bf7b46'

class ProviderUnavailableError(ValueError):
    pass


class TokenProvider(abc.ABC):
    @abc.abstractmethod
    def available(self):
        # type: () -> bool
        raise NotImplementedError()

    @abc.abstractmethod
    def get_token(self, scopes=None, username=None):
        # type: (*str) -> {str:str}
        raise NotImplementedError()


class TokenProviderChain(TokenProvider):
    def __init__(self, *args):
        self._links = list(args)

    def available(self):
        return any((item for item in self._links if item.available()))

    def get_token(self, scopes=None, username=None):
        return next((item.get_token(scopes=scopes) for item in self._links if item.available()))


class SharedTokenCacheProvider(TokenProvider):

    def __init__(self, client_id=None, cache_location=None):
        client_id = client_id or _DEFAULT_CLIENT_ID
        token_cache = get_protected_token_cache(cache_location=cache_location)
        self._app = PublicClientApplication(client_id=client_id, token_cache=token_cache)

    def available(self):
        return any(self._get_accounts())

    def get_token(self, scopes=None, username=None):
        accounts = self._get_accounts(username=username)
        if any(accounts):
            active_account = accounts[0]
            return self._app.acquire_token_silent(scopes=scopes, account=active_account)
        raise ProviderUnavailableError()

    def _get_accounts(self, username=None):
        return self._app.get_accounts(username=username)


class ServicePrincipalProvider(TokenProvider):

    def __init__(self, client_id=None, client_secret=None):
        client_id = client_id or os.getenv('AZURE_CLIENT_ID')
        client_secret = client_secret or os.getenv('AZURE_CLIENT_SECRET')

        self._app = ConfidentialClientApplication(
            client_id=client_id,
            client_secret=client_secret,
        )

    def available(self):
        """ Always returns true, because if it was able to be instantiated, it is available for use."""
        return True

    def get_token(self, scopes=None):
        return self._app.acquire_token_for_client(scopes=scopes)


DEFAULT_TOKEN_CHAIN = TokenProviderChain(
    SharedTokenCacheProvider())



#!/usr/bin/env python

from setuptools import setup, find_packages

__version__ = "0.0.4"

setup(
    name='marstr-msal-extensions',
    version=__version__,
    package_dir={
        '': 'src',
    },
    packages=find_packages(
        where="./src",
    ),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
    ],
    install_requires=[
        'msal~=0.3.0',
        'psutil~=5.6',
        'portalocker~=1.4',
    ],
    extras_require={
        'dev': [
            'pytest',
        ]
    },
)
